#include <windows.h>
#include "lightmap.h"
#include "collision.h"
#include "bsp.h"
#include "texture.h"
#include "tga.h"
#include "mmgr.h"

extern int numStaticLights;
extern char AppDirectory[MAX_PATH];
extern int numLightmaps;
extern StaticLight* staticlight;

void SetStaticLights(StaticLight* staticlight)
{
    staticlight[0].Position.x = -4.0;
    staticlight[0].Position.y = 9.0;
    staticlight[0].Position.z = -50.0;
    staticlight[0].Radius = 300.0;
    staticlight[0].Brightness = 5.0;
    staticlight[0].Red = 255.0;
    staticlight[0].Green = 255.0;
    staticlight[0].Blue = 255.0;

    staticlight[1].Position.x = -28.0;
    staticlight[1].Position.y = 9.0;
    staticlight[1].Position.z = -50.0;
    staticlight[1].Radius = 300.0;
    staticlight[1].Brightness = 5.0;
    staticlight[1].Red = 255.0;
    staticlight[1].Green = 255.0;
    staticlight[1].Blue = 255.0;

    staticlight[2].Position.x = -52.0;
    staticlight[2].Position.y = 9.0;
    staticlight[2].Position.z = -50.0;
    staticlight[2].Radius = 300.0;
    staticlight[2].Brightness = 5.0;
    staticlight[2].Red = 255.0;
    staticlight[2].Green = 255.0;
    staticlight[2].Blue = 255.0;

    staticlight[3].Position.x = -4.0;
    staticlight[3].Position.y = 9.0;
    staticlight[3].Position.z = 50.0;
    staticlight[3].Radius = 300.0;
    staticlight[3].Brightness = 5.0;
    staticlight[3].Red = 255.0;
    staticlight[3].Green = 255.0;
    staticlight[3].Blue = 255.0;

    staticlight[4].Position.x = -28.0;
    staticlight[4].Position.y = 9.0;
    staticlight[4].Position.z = 50.0;
    staticlight[4].Radius = 300.0;
    staticlight[4].Brightness = 5.0;
    staticlight[4].Red = 255.0;
    staticlight[4].Green = 255.0;
    staticlight[4].Blue = 255.0;

    staticlight[5].Position.x = -52.0;
    staticlight[5].Position.y = 9.0;
    staticlight[5].Position.z = 50.0;
    staticlight[5].Radius = 300.0;
    staticlight[5].Brightness = 5.0;
    staticlight[5].Red = 255.0;
    staticlight[5].Green = 255.0;
    staticlight[5].Blue = 255.0;

    staticlight[6].Position.x = 110.0;
    staticlight[6].Position.y = 5.0;
    staticlight[6].Position.z = 120.0;
    staticlight[6].Radius = 100.0;
    staticlight[6].Brightness = 30.0;
    staticlight[6].Red = 255.0;
    staticlight[6].Green = 255.0;
    staticlight[6].Blue = 255.0;

    staticlight[7].Position.x = 140.0;
    staticlight[7].Position.y = 0.0;
    staticlight[7].Position.z = -20.0;
    staticlight[7].Radius = 100.0;
    staticlight[7].Brightness = 10.0;
    staticlight[7].Red = 0.0;
    staticlight[7].Green = 255.0;
    staticlight[7].Blue = 255.0;

    staticlight[8].Position.x = 0.0;
    staticlight[8].Position.y = 25.0;
    staticlight[8].Position.z = 0.0;
    staticlight[8].Radius = 45.0;
    staticlight[8].Brightness = 12.0;
    staticlight[8].Red = 250.0;
    staticlight[8].Green = 170.0;
    staticlight[8].Blue = 50.0;
}

void CreateLightmaps(int numpolys, POLYGON* polylist, Lightmap* lightmaplist)
{
    const int Width = 16;
    const int Height = 16;
    VECTOR lumels[Width][Height];               // world coordinates of lumels
    VECTOR edge1, edge2;
    VECTOR newedge1, newedge2;
    VECTOR Vect1, Vect2;
    VECTOR UVVector;
    VECTOR poly_normal;
    VECTOR lightvector;
    VECTOR pointonplane;
    float lumelcolor[Width][Height][3];  // array to hold lumel rgb values
    float lightdistance;
    float uvMin_U;
    float uvMin_V;
    float uvMax_U;
    float uvMax_V;
    float cosAngle;
    float X, Y, Z;
    float Distance;
    float uvDelta_U;
    float uvDelta_V;
    float ufactor;
    float vfactor;
    float combinedred;
    float combinedgreen;
    float combinedblue;
    float intensity;
    int flag; // set to 1, 2 or 3 depending on whether normal's X, Y or Z component is largest
    char temp[256];
    unsigned char* lightmap = new unsigned char[Width * Height * 4];
        
    for (int loop = 0; loop < numpolys; loop++)
    {
        // get uv's in 3D worldspace
        poly_normal = polylist[loop].GetNormal();
        poly_normal.Normalize();
        pointonplane.x = polylist[loop].Vertex[0].x;
        pointonplane.y = polylist[loop].Vertex[0].y;
        pointonplane.z = polylist[loop].Vertex[0].z;

        // find plane to map onto
        if (fabs(poly_normal.x) > fabs(poly_normal.y) && fabs(poly_normal.x) > fabs(poly_normal.z))
        {
            flag = 1;
            lightmaplist[loop].vertex_u[0] = polylist[loop].Vertex[0].y;
            lightmaplist[loop].vertex_v[0] = polylist[loop].Vertex[0].z;
            lightmaplist[loop].vertex_u[1] = polylist[loop].Vertex[1].y;
            lightmaplist[loop].vertex_v[1] = polylist[loop].Vertex[1].z;
            lightmaplist[loop].vertex_u[2] = polylist[loop].Vertex[2].y;
            lightmaplist[loop].vertex_v[2] = polylist[loop].Vertex[2].z;
        }
        else if (fabs(poly_normal.y) > fabs(poly_normal.x) && fabs(poly_normal.y) > fabs(poly_normal.z))
        {
            flag = 2;
            lightmaplist[loop].vertex_u[0] = polylist[loop].Vertex[0].x;
            lightmaplist[loop].vertex_v[0] = polylist[loop].Vertex[0].z;
            lightmaplist[loop].vertex_u[1] = polylist[loop].Vertex[1].x;
            lightmaplist[loop].vertex_v[1] = polylist[loop].Vertex[1].z;
            lightmaplist[loop].vertex_u[2] = polylist[loop].Vertex[2].x;
            lightmaplist[loop].vertex_v[2] = polylist[loop].Vertex[2].z;
        }
        else
        {
            flag = 3;
            lightmaplist[loop].vertex_u[0] = polylist[loop].Vertex[0].x;
            lightmaplist[loop].vertex_v[0] = polylist[loop].Vertex[0].y;
            lightmaplist[loop].vertex_u[1] = polylist[loop].Vertex[1].x;
            lightmaplist[loop].vertex_v[1] = polylist[loop].Vertex[1].y;
            lightmaplist[loop].vertex_u[2] = polylist[loop].Vertex[2].x;
            lightmaplist[loop].vertex_v[2] = polylist[loop].Vertex[2].y;
        }

        // convert to 2D texture space
        uvMin_U = lightmaplist[loop].vertex_u[0];
        uvMin_V = lightmaplist[loop].vertex_v[0];
        uvMax_U = lightmaplist[loop].vertex_u[0];
        uvMax_V = lightmaplist[loop].vertex_v[0];

        for (int i = 0; i < 3; i++)
        {
            if (lightmaplist[loop].vertex_u[i] < uvMin_U )
                uvMin_U = lightmaplist[loop].vertex_u[i];
            if (lightmaplist[loop].vertex_v[i] < uvMin_V )
                uvMin_V = lightmaplist[loop].vertex_v[i];
            if (lightmaplist[loop].vertex_u[i] > uvMax_U )
                uvMax_U = lightmaplist[loop].vertex_u[i];
            if (lightmaplist[loop].vertex_v[i] > uvMax_V )
                uvMax_V = lightmaplist[loop].vertex_v[i];
        }

        uvDelta_U = uvMax_U - uvMin_U;
        uvDelta_V = uvMax_V - uvMin_V;

        for (int i = 0; i < 3; i++)
        {
            lightmaplist[loop].vertex_u[i] -= uvMin_U;
            lightmaplist[loop].vertex_v[i] -= uvMin_V;
            lightmaplist[loop].vertex_u[i] /= uvDelta_U;
            lightmaplist[loop].vertex_v[i] /= uvDelta_V;
        }

        // calculate the world space position for each lumel
        Distance = - (poly_normal.x * pointonplane.x + poly_normal.y * pointonplane.y + poly_normal.z * pointonplane.z);

        switch (flag)
        {
            case 1: //YZ Plane
                X = - ( poly_normal.y * uvMin_U + poly_normal.z * uvMin_V + Distance ) / poly_normal.x;
                UVVector.x = X;
                UVVector.y = uvMin_U;
                UVVector.z = uvMin_V;
                X = - ( poly_normal.y * uvMax_U + poly_normal.z * uvMin_V + Distance ) / poly_normal.x;
                Vect1.x = X;
                Vect1.y = uvMax_U;
                Vect1.z = uvMin_V;
                X = - ( poly_normal.y * uvMin_U + poly_normal.z * uvMax_V + Distance ) / poly_normal.x;
                Vect2.x = X;
                Vect2.y = uvMin_U;
                Vect2.z = uvMax_V;
            break;

            case 2: //XZ Plane
                Y = - ( poly_normal.x * uvMin_U + poly_normal.z * uvMin_V + Distance ) / poly_normal.y;
                UVVector.x = uvMin_U;
                UVVector.y = Y;
                UVVector.z = uvMin_V;
                Y = - ( poly_normal.x * uvMax_U + poly_normal.z * uvMin_V + Distance ) / poly_normal.y;
                Vect1.x = uvMax_U;
                Vect1.y = Y;
                Vect1.z = uvMin_V;
                Y = - ( poly_normal.x * uvMin_U + poly_normal.z * uvMax_V + Distance ) / poly_normal.y;
                Vect2.x = uvMin_U;
                Vect2.y = Y;
                Vect2.z = uvMax_V;
            break;

            case 3: //XY Plane
                Z = - ( poly_normal.x * uvMin_U + poly_normal.y * uvMin_V + Distance ) / poly_normal.z;
                UVVector.x = uvMin_U;
                UVVector.y = uvMin_V;
                UVVector.z = Z;
                Z = - ( poly_normal.x * uvMax_U + poly_normal.y * uvMin_V + Distance ) / poly_normal.z;
                Vect1.x = uvMax_U;
                Vect1.y = uvMin_V;
                Vect1.z = Z;
                Z = - ( poly_normal.x * uvMin_U + poly_normal.y * uvMax_V + Distance ) / poly_normal.z;
                Vect2.x = uvMin_U;
                Vect2.y = uvMax_V;
                Vect2.z = Z;
            break;
        }

        edge1.x = Vect1.x - UVVector.x;
        edge1.y = Vect1.y - UVVector.y;
        edge1.z = Vect1.z - UVVector.z;
        edge2.x = Vect2.x - UVVector.x;
        edge2.y = Vect2.y - UVVector.y;
        edge2.z = Vect2.z - UVVector.z;
        for(int iX = 0; iX < Width; iX++)
        {
            for(int iY = 0; iY < Height; iY++)
            {
                ufactor = (iX / (float)Width);
                vfactor = (iY / (float)Height);
                newedge1.x = edge1.x * ufactor;
                newedge1.y = edge1.y * ufactor;
                newedge1.z = edge1.z * ufactor;
                newedge2.x = edge2.x * vfactor;
                newedge2.y = edge2.y * vfactor;
                newedge2.z = edge2.z * vfactor;

                lumels[iX][iY].x = UVVector.x + newedge2.x + newedge1.x;
                lumels[iX][iY].y = UVVector.y + newedge2.y + newedge1.y;
                lumels[iX][iY].z = UVVector.z + newedge2.z + newedge1.z;

                combinedred = 0.0;
                combinedgreen = 0.0;
                combinedblue = 0.0;
                for (int i = 0; i < numStaticLights; i++)
                {
                    if (ClassifyPoint(staticlight[i].Position, pointonplane, poly_normal) == 1)
                    {
                        lightvector.x = staticlight[i].Position.x - lumels[iX][iY].x;
                        lightvector.y = staticlight[i].Position.y - lumels[iX][iY].y;
                        lightvector.z = staticlight[i].Position.z - lumels[iX][iY].z;
                        lightdistance = lightvector.GetMagnitude();
                        lightvector.Normalize();
                        cosAngle = DotProduct(poly_normal, lightvector);
                        if (lightdistance < staticlight[i].Radius)
                        {
                            intensity = (staticlight[i].Brightness * cosAngle) / lightdistance;
                            combinedred += staticlight[i].Red * intensity;
                            combinedgreen += staticlight[i].Green * intensity;
                            combinedblue += staticlight[i].Blue * intensity;
                        }
                    }
                }
                if (combinedred > 255.0)
                    combinedred = 255.0;
                if (combinedgreen > 255.0)
                    combinedgreen = 255.0;
                if (combinedblue > 255.0)
                    combinedblue = 255.0;
                lumelcolor[iX][iY][0] = combinedred;
                lumelcolor[iX][iY][1] = combinedgreen;
                lumelcolor[iX][iY][2] = combinedblue;
            }
        }

        for(int iX = 0; iX < (Width * 4); iX += 4)
        {
            for(int iY = 0; iY < Height; iY += 1)
            {
                lightmap[iX + iY * Height * 4] = (char)lumelcolor[iX / 4][iY][0];         // Red
                lightmap[iX + iY * Height * 4 + 1] = (char)lumelcolor[iX / 4][iY][1];     // Green
                lightmap[iX + iY * Height * 4 + 2] = (char)lumelcolor[iX / 4][iY][2];     // Blue
                lightmap[iX + iY * Height * 4 + 3] = 255;                                 // Alpha
            }
        }

        sprintf(lightmaplist[loop].Texture.TexName, "%s", AppDirectory);
        strcat(lightmaplist[loop].Texture.TexName, "\\images\\lm");
        strcat(lightmaplist[loop].Texture.TexName, itoa(numLightmaps++, temp, 10));
        strcat(lightmaplist[loop].Texture.TexName, ".tga");
        tgaSave(lightmaplist[loop].Texture.TexName, Width, Height, 32, lightmap);
    }
    delete[] lightmap;
    return;
}

void CreateBSPLightmaps(BSP_node* node)
{
    if (node->leaf)
    {
        node->nodelightmaplist = new Lightmap[node->numpolys];
        CreateLightmaps(node->numpolys, node->nodepolylist, node->nodelightmaplist);
        return;
    }
    CreateBSPLightmaps(node->frontnode);
    CreateBSPLightmaps(node->backnode);
}

void LoadBSPLightmaps(BSP_node* node)
{
    if (node->leaf)
    {
    for (int i = 0; i < node->numpolys; i++)
    {
        node->nodelightmaplist[i].Texture.LoadTGA(GL_LINEAR, GL_NEAREST_MIPMAP_LINEAR/*GL_NEAREST, GL_NEAREST*/, GL_CLAMP, GL_CLAMP, 1);
                
    }
    return;
    }
    LoadBSPLightmaps(node->frontnode);
    LoadBSPLightmaps(node->backnode);
}

